package cz.upol.fpj.lecture12.collections

abstract class Tree[K,V] {
  def put(k: K, v: V): Tree[K,V]
  def print(level: Int = 0)
  def foreach(f: (K,V) => Unit)
  def size(): Int
  def get(k: K): Option[V]
}

case class EmptyTree[K,V]()(implicit cmp: Ordering[K]) extends Tree[K,V] {
  def put(k: K, v: V) = Node(k, v, EmptyTree(), EmptyTree())
  def print(level: Int = 0) = {}
  def foreach(f: (K,V) => Unit) = {}
  def size() = 0
  def get(k: K) = None
}

case class Node[K,V](
    val key: K,
    val value: V,
    left: Tree[K,V],
    right: Tree[K,V])(implicit cmp: Ordering[K]) extends Tree[K,V] {
  
  def put(k: K, v: V) = cmp.compare(k, key) match {
    case 0 => Node(k, v, left, right)
    case x if (x < 0) => Node(key, value, left.put(k, v), right)
    case _ => Node(key, value, left, right.put(k, v))
  }
  
  def get(k: K) = cmp.compare(k, key) match {
    case 0 => Some(value)
    case x if (x < 0) => left.get(k)
    case _ => right.get(k)
  }
  
  def foreach(f: (K,V) => Unit) = {
    left.foreach(f)
    f(key, value)
    right.foreach(f)
  }
  
  def size() = 1 + left.size + right.size
    
  def print(level: Int) = {
    left.print(level + 1)
    val indent = (for (i <- 0 to (level - 1)) yield "  ").mkString
    println (indent + key + " = " + value)
    right.print(level + 1)
  }
}


object Tree {
  
  def apply[K,V](assoc: (K, V)*)(implicit cmp: Ordering[K]) = {
    val empty: Tree[K,V] = EmptyTree[K, V]()
    assoc.foldLeft(empty)((t, a) => t.put(a._1, a._2))
  }

  def main(args: Array[String]): Unit = {
//    val t = EmptyTree[Int, String]().put(10, "aaa").put(20, "ccc").put(5, "bbb")
    val t = Tree(10 -> "aaa", 20 -> "ccc", 5 -> "bbb")
    t.print()
  }
    
}