package cz.upol.fpj.lecture12.implicits

abstract class Tree[K,V] {
  def put(k: K, v: V): Tree[K,V]
  def print(level: Int = 0)
}

case class EmptyTree[K,V]()(implicit cmp: (K,K) => Int) extends Tree[K,V] {
  def put(k: K, v: V) = Node(k, v, EmptyTree(), EmptyTree())
  def print(level: Int = 0) = {}
}

case class Node[K,V](
    val key: K,
    val value: V,
    left: Tree[K,V],
    right: Tree[K,V])(implicit cmp: (K,K) => Int) extends Tree[K,V] {
  
  def put(k: K, v: V) = cmp(key, k) match {
    case 0 => Node(k, v, left, right)
    case x if (x < 0) => Node(key, value, left.put(k, v), right)
    case _ => Node(key, value, left, right.put(k, v))
  }
  
    
  def print(level: Int) = {
    left.print(level + 1)
    val indent = (for (i <- 0 to (level - 1)) yield "  ").mkString
    println (indent + key + " = " + value)
    right.print(level + 1)
  }
}


object Tree {
  
  implicit val cmpInteger: (Int, Int) => Int = (_ - _)
  
  def apply[K,V](assoc: (K, V)*)(implicit cmp: (K,K) => Int) = {
    val empty: Tree[K,V] = EmptyTree[K, V]()
    assoc.foldLeft(empty)((t, a) => t.put(a._1, a._2))
  }

  def main(args: Array[String]): Unit = {
//    val t = EmptyTree[Int, String]().put(10, "aaa").put(20, "ccc").put(5, "bbb")
//    val t = Tree(10 -> "aaa", 20 -> "ccc", 5 -> "bbb")(_ - _)
//    t.print()
  }
    
}