package cz.upol.fpj.lecture12.implicits

abstract class TreeO[K <: Comparable[K],V] {
  def put(k: K, v: V): TreeO[K,V]
  def print(level: Int = 0)
}

case class EmptyTreeO[K <: Comparable[K],V]() extends TreeO[K,V] {
  def put(k: K, v: V) = NodeO(k, v, EmptyTreeO(), EmptyTreeO())
  def print(level: Int = 0) = {}
}

case class NodeO[K <: Comparable[K], V](
    val key: K,
    val value: V,
    left: TreeO[K,V],
    right: TreeO[K,V]) extends TreeO[K,V] {
  
  def put(k: K, v: V) = key compareTo k  match {
    case 0 => NodeO(k, v, left, right)
    case x if (x < 0) => NodeO(key, value, left.put(k, v), right)
    case _ => NodeO(key, value, left, right.put(k, v))
  }
  
    
  def print(level: Int) = {
    left.print(level + 1)
    val indent = (for (i <- 0 to (level - 1)) yield "  ").mkString
    println (indent + key + " = " + value)
    right.print(level + 1)
  }
}


object TreeO {
  
  def main(args: Array[String]): Unit = {
     val t = EmptyTreeO[Integer, String]().put(10, "aaa").put(20, "ccc").put(5, "bbb")
     t.print()
  }
    
}