package cz.upol.fpj.lecture12.prolog

import scala.collection.mutable.ArrayBuffer

class Prolog {
  /** pomocna promenna urcena ke generovani unikatnich instancni promennych */
  private var varIndex = 0;
  
  /** kolekce pravidel */
  private val rules = new ArrayBuffer[Rule];
  
  /**
   * Vlozi pravidlo
   */
  def add(head: Term, args: Term*) = rules += new Rule(head, args);
  
  /**
   * Polozi dotaz
   */
  def query(t: Term) = {
    val q = t.createInstance(varIndex);
    val subst = resolve(Set(q), Map())
    
    // seznam vsech promennych v dotazu
    val vars = extractVarInst(q)
    
    // vybere substituce odpovidajici danemu dotazu
    vars.map(x => (x.name, subst.get(x)))
  }
  
  /**
   * Snazi se najit vhodnou substituci pro termy v kolekce workingSet
   */
  def resolve(workingSet: Set[Term], subst: Map[VarInst, Term]): Option[Map[VarInst, Term]] = {
    
    // pokude je workingSet prazdny konci
    if (workingSet.isEmpty) return Some(subst);

    varIndex += 1
    
    // prochazi vsechna pravidla a termy ve workingSet
    for (
      r <- rules map { x => x.createInstance(varIndex) };
      w <- workingSet
    ) {
        // zkusi na term "w" pouzit pravidlo "r" a pokracuje v rezoluci
        val solution = tryRule(w, r, workingSet, subst);
        
        // pokud existuje reseni, vrati jej
        if (solution.isDefined) return solution;
    };
    return None;
  }

  private def tryRule(t: Term, r: Rule, workingSet: Set[Term], subst: Map[VarInst, Term]): Option[Map[VarInst, Term]] = {
    Term.unify(r.head, t).map { u =>
      // odstrani term, prida telo pravidla r a provede substituce
      val newWorkingSet = ((workingSet - t) ++ r.body) map (_ subst u) 
      
      // provede substituce v substitucich
      val newSubst = u ++ (subst map { case (key, value) => (key, value.subst(u)) })
      
      // pokracuje v rezoluci
      resolve(newWorkingSet, newSubst)
    }.flatten
  }

  /**
   * Vraci mnozinu vsech objektu VarInst, ktere se v termu vyskytuji
   */
  private def extractVarInst(t: Term): Set[VarInst] = t match {
    case t: VarInst => Set(t);
    case CompoundTerm(name, args) => args.flatMap(x => extractVarInst(x)).toSet
    case _ => Set.empty;
  }
}