package cz.upol.fpj.lecture12.prolog

/**
 * abstraktni trida popisujici jednotlive termy
 */
abstract class Term {
  def subst(s: Map[VarInst,Term]): Term
  def createInstance(id: Int): Term
}

object Term {
    
  /** 
   * Zkusi unifikovat dany termy t1 a t2,
   * pokud takova unifikace neexistuje vraci None,
   * jinak vraci kolekci substituci typu promenna->term
   */
  def unify(t1: Term, t2: Term): Option[Map[VarInst,Term]] = (t1, t2) match  {
    case (a: VarInst, b: Term) => Some(Map(a -> b))
    case (a: Term, b: VarInst) => Some(Map(b -> a))
    case (Atom(a), Atom(b)) => if (a.equals(b)) Some(Map.empty) else None;
    case (CompoundTerm(a, args1), CompoundTerm(b, args2)) => {
      
      // pokud se shoduje hlava termu
      if (a.equals(b)) { 
        
        // unifiuje jednotlive podtermy
        val args = (args1 zip args2) map { x => unify(x._1, x._2) }
        
        // pokud nektery z podtermu neni unifikovatelny vraci None, jinak je spoji
        if (args.contains(None)) None else Some(merge(args.flatten)); 
      } else None;
    }
    case _ => None;
  };
  
    
  private def merge(maps: Seq[Map[VarInst,Term]]) = 
    maps.foldLeft(Map[VarInst,Term]()) { (a, b) => a ++ b};
  
}

case class Atom(value: Any) extends Term {
  override def subst(s: Map[VarInst,Term]) = this;
  override def createInstance(id: Int) = this;
  override def toString = value.toString
}
  
case class Var(name: String) extends Term {
  override def subst(s: Map[VarInst,Term]) = this
  override def createInstance(id: Int) = VarInst(name, id);
  override def toString = name.toString
}

case class VarInst(name: String, instance: Int) extends Term {
  override def subst(s: Map[VarInst,Term]) = s.get(this).getOrElse(this);
  override def createInstance(id: Int) = throw new Exception("Error!");
  override def toString = name.toString + "/" + instance
}

case class CompoundTerm(name: String, args: Seq[Term]) extends Term {
  override def subst(s: Map[VarInst,Term]) = CompoundTerm(name, args.map(x => x.subst(s)));
  override def createInstance(id: Int) = CompoundTerm(name, args.map(x => x.createInstance(id)))
  override def toString = name + "(" + args.mkString(",") + ")"
}

class Rule(val head: Term, val body: Seq[Term]) {
  def createInstance(id: Int) = new Rule(head.createInstance(id), body map { t => t.createInstance(id) })
}